from time import monotonic_ns
from typing import Union

import jax
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from etils.array_types import PRNGKey
from jax import numpy as jnp, random, tree_map
from jax import vmap
from jax.lax import dynamic_update_slice

from jaxns import TerminationCondition, Reservoir, ApproximateNestedSampler, Model, Prior, PriorModelGen
from jaxns.internals.log_semiring import LogSpace
from jaxns.internals.maps import replace_index
from jaxns.static_uniform import BadUniformSampler
from jaxns.statistics import compute_num_live_points_from_unit_threads, compute_remaining_evidence, \
    perfect_live_point_computation_jax, fast_perfect_live_point_computation_jax, fast_triu_rowsum, combine_reservoirs

tfpd = tfp.distributions


def _sure_compute_num_live_points_from_unit_threads(log_L_constraints, log_L_samples, num_samples=None, debug=False):
    if num_samples is None:
        num_samples = log_L_samples.size

    empty_mask = jnp.arange(log_L_samples.size) >= num_samples
    log_L_samples = jnp.where(empty_mask, jnp.inf, log_L_samples)
    idx_sort = jnp.lexsort((log_L_constraints, log_L_samples))
    empty_mask = empty_mask[idx_sort]
    num_live_points = []

    available = jnp.bitwise_not(empty_mask)  # jnp.ones(log_L_samples.shape, jnp.bool_)
    log_L_samples = log_L_samples[idx_sort]
    log_L_constraints = log_L_constraints[idx_sort]
    contour = log_L_constraints[0]

    if debug:
        print("###")
        print(f"log_L_constraints: {log_L_constraints}")
        print(f"log_L_samples: {log_L_samples}")
    for idx in range(log_L_samples.size):
        # by definition the number of live points is count of points where all three conditions are true:
        # 1) not yet consumed (once consumed no longer valid)
        # 2) sampling constraint of sample are <= contour (uniform sampling within a smaller contour also uniformly samples within a larger contour)
        # 3) log(L) of sample >= current contour (we shrink upwards in likelihood, the equality makes this work for plateaus)
        mask = available & (log_L_samples >= contour) & (log_L_constraints <= contour)
        n = jnp.sum(mask)
        # n = jnp.sum(available & (log_L_constraints <= contour)) # also valid because we sort samples
        if debug:
            print("---")
            print(f"L*: {contour} n:{n}")
            for j in range(log_L_samples.size):
                s = f"{'*' if mask[j] else ' '}\t{j}: available: [{'x' if available[j] else ''}] " \
                    f"log(L)>contour ({log_L_samples[j]},{contour}): [{'x' if log_L_samples[j] > contour else ''}] " \
                    f"L*<=contour ({log_L_constraints[j]},{contour}): [{'x' if log_L_constraints[j] <= contour else ''}]"
                print(s)

        contour = log_L_samples[idx]
        available = dynamic_update_slice(available, jnp.asarray([False]), [idx])
        num_live_points.append(n)

    num_live_points = jnp.stack(num_live_points)

    return num_live_points


def perfect_live_point_computation(log_L_constraints, log_L_samples, num_samples: Union[int, None] = None, debug=True):
    log_L_constraints = np.array(log_L_constraints)
    log_L_samples = np.array(log_L_samples)
    sort_idx = np.lexsort((log_L_constraints, log_L_samples))

    log_L_contour = log_L_constraints[sort_idx[0]]

    num_live_points = []
    idx = 0
    while idx < log_L_samples.size:
        log_L_dead = log_L_samples[sort_idx[idx]]
        while idx < log_L_samples.size:
            if log_L_samples[sort_idx[idx]] != log_L_dead:
                break
            count = np.sum(np.bitwise_and(log_L_contour < log_L_samples, log_L_constraints <= log_L_contour))
            if debug:
                print('index', sort_idx[idx], 'dead point', log_L_dead, count, ' points within', log_L_contour)
            log_L_samples[sort_idx[idx]] = -np.inf
            log_L_constraints[sort_idx[idx]] = np.inf
            num_live_points.append(count)
            idx += 1
        log_L_contour = log_L_dead
    if debug:
        print(num_live_points)
    num_live_points = jnp.asarray(num_live_points)
    if num_samples is not None:
        empty_mask = np.greater_equal(np.arange(log_L_samples.size), num_samples)
        num_live_points = np.where(empty_mask, 0, num_live_points)
    return num_live_points, sort_idx


def test_perfect_live_point_computation_jax():
    log_L_constraints = random.uniform(random.PRNGKey(42), shape=(100,))
    log_L_samples = random.uniform(random.PRNGKey(43), shape=(100,), minval=log_L_constraints)
    num_live_points, sort_idx = perfect_live_point_computation(log_L_constraints, log_L_samples)
    num_live_points_jax, sort_idx_jax = perfect_live_point_computation_jax(log_L_constraints, log_L_samples)
    assert jnp.all(num_live_points_jax == num_live_points)


def test_fast_perfect_live_point_computation_jax():
    for i in range(10):
        log_L_constraints = random.uniform(random.PRNGKey(i), shape=(100,))
        log_L_samples = random.uniform(random.PRNGKey(i + 100000), shape=(100,), minval=log_L_constraints)
        num_live_points, sort_idx = perfect_live_point_computation(log_L_constraints, log_L_samples, debug=False)
        num_live_points_jax, sort_idx_jax = perfect_live_point_computation_jax(log_L_constraints, log_L_samples)
        num_live_points_fast, sort_idx = fast_perfect_live_point_computation_jax(log_L_constraints, log_L_samples)
        assert jnp.all(num_live_points_jax == num_live_points)
        assert jnp.all(num_live_points_fast == num_live_points)

    for n in [1e3, 1e4, 2e4]:
        f1 = jax.jit(perfect_live_point_computation_jax)
        f2 = jax.jit(fast_perfect_live_point_computation_jax)
        log_L_constraints = random.uniform(random.PRNGKey(42), shape=(int(n),))
        log_L_samples = random.uniform(random.PRNGKey(420), shape=(int(n),), minval=log_L_constraints)
        num_live_points, _ = f1(log_L_constraints, log_L_samples)
        num_live_points.block_until_ready()
        t0 = monotonic_ns()
        for _ in range(10):
            num_live_points, _ = f1(log_L_constraints, log_L_samples)
            num_live_points.block_until_ready()
        dt_normal = (monotonic_ns() - t0) / 1e9 / 10
        print(f"Normal {n}: {dt_normal} sec")

        num_live_points, _ = f2(log_L_constraints, log_L_samples)
        num_live_points.block_until_ready()
        t0 = monotonic_ns()
        for _ in range(10):
            num_live_points, _ = f2(log_L_constraints, log_L_samples)
            num_live_points.block_until_ready()
        dt_fast = (monotonic_ns() - t0) / 1e9 / 10
        print(f"Fast {n}: {dt_fast} sec")

        assert dt_fast < dt_normal

    # ensure scaling law is sub linear
    n_array = [1e3, 1e4, 1e5, 1e6]
    dt_array = []
    for n in n_array:
        num_live_points, _ = f2(log_L_constraints, log_L_samples)
        num_live_points.block_until_ready()
        t0 = monotonic_ns()
        for _ in range(10):
            num_live_points, _ = f2(log_L_constraints, log_L_samples)
            num_live_points.block_until_ready()
        dt_fast = (monotonic_ns() - t0) / 1e9 / 10
        print(f"Fast {n}: {dt_fast} sec")
        dt_array.append(dt_fast)

    def is_sub_linear(n_array, dt_array):
        if len(n_array) != len(dt_array):
            raise ValueError("Both arrays must have the same length.")

        if len(n_array) < 2:
            raise ValueError("At least two data points are needed to determine the scaling law.")

        sub_linear = True

        for i in range(1, len(n_array)):
            n_ratio = n_array[i] / n_array[i - 1]
            dt_ratio = dt_array[i] / dt_array[i - 1]
            print(n_ratio, dt_ratio)

            # If the ratio of execution times is greater than or equal to the ratio of input sizes,
            # it is not sub-linear.
            if dt_ratio >= n_ratio:
                sub_linear = False
                break

        return sub_linear

    assert is_sub_linear(n_array, dt_array)


def _sure_infimum_constraint(log_L_constraints, log_L_samples, sort_idx):
    """
    For a single sample find `i` such that log_L_contours[i] is the greatest strict lower bound of log_L_sample.

    Args:
        log_L_constraints:
        log_L_samples:
        sort_idx:
    """
    log_L_constraints = log_L_constraints[sort_idx]
    log_L_samples = log_L_samples[sort_idx]
    log_L_contours = jnp.concatenate([log_L_constraints[0:1], log_L_samples[:-1]], axis=0)
    print("###")
    print(f'log_L_contours:{log_L_contours}')
    print(f'log_L_samples:{log_L_samples}')
    indices = []
    indices = jnp.arange(log_L_contours.size)
    for _ in range(log_L_samples.size):
        # indices = jnp.where((log_L_contours[indices]==log_L_contours[indices-1]) | (log_L_contours[indices] == log_L_samples), indices-1,indices)
        indices = jnp.clip(jnp.where((log_L_contours[indices] == log_L_samples), indices - 1, indices), -1,
                           log_L_samples.size - 1)

    print(f"indices:{indices}")
    return jnp.asarray(indices)


def test_fast_triu_rowsum():
    def exact(a, b):
        M = a[:, None] * b[None, :]
        res = []
        for i in range(a.size):
            res.append(np.sum(M[i, i:]))
        return np.asarray(res)

    a = np.array([1, 2, 3, 4])
    b = np.array([5, 6, 7, 8])

    assert np.allclose(fast_triu_rowsum(a, b), exact(a, b))


def test_fast_sum_inequality():
    def exact(a, b):
        res = []
        for i in range(a.size):
            res.append(jnp.sum(a[i] <= b[i:]))
        return jnp.asarray(res)

    def upper_triangle_sum(a, b):
        b_minus_cumsum_a = b - np.cumsum(a)
        count_positive = np.cumsum(np.concatenate(([0], b_minus_cumsum_a[:-1])) >= 0)
        result = count_positive - np.concatenate(([0], count_positive[:-1]))
        return result

    a = np.array([1, 2, 4, 5])
    b = np.array([3, 4, 5, 6])

    print(exact(a, b))
    print(upper_triangle_sum(a, b))


def test_compute_num_live_points():
    from jax import random

    def run_test(log_L_constraints, log_L_samples, num_live_points, num_samples=None, debug=False):
        print("log_L_samples", log_L_samples, 'log_L_constraints', log_L_constraints)
        num_live_points_perfect, sort_idx_perfect = perfect_live_point_computation(log_L_constraints, log_L_samples,
                                                                                   num_samples=num_samples)
        assert np.all(num_live_points_perfect == num_live_points)
        print('1 passed')

        num_live_points_perfect_jax, sort_idx_perfect_jax = perfect_live_point_computation_jax(log_L_constraints,
                                                                                               log_L_samples,
                                                                                               num_samples=num_samples)
        # assert np.all(num_live_points_perfect_jax == num_live_points)
        # print('2 passed')

        num_live_points_perfect_jax_fast, sort_idx_perfect_jax_fast = fast_perfect_live_point_computation_jax(
            log_L_constraints,
            log_L_samples,
            num_samples=num_samples)
        assert np.all(num_live_points_perfect_jax_fast == num_live_points)
        print('3 passed')

        # assert jnp.allclose(num_live_points,
        #                     _sure_compute_num_live_points_from_unit_threads(log_L_constraints, log_L_samples,
        #                                                                     num_samples, debug=False))

        # test_num_live_points, sort_idx = compute_num_live_points_from_unit_threads(log_L_constraints, log_L_samples,
        #                                                                            num_samples, sorted_collection=False)
        # assert jnp.all(num_live_points_perfect == test_num_live_points)
        # assert jnp.all(infimum_constraint(log_L_constraints, log_L_samples, sort_idx) == _sure_infimum_constraint(
        #     log_L_constraints, log_L_samples, sort_idx))

    # constraints are previous samples
    log_L_constraints = jnp.asarray([0., 0.5, 2.])
    log_L_samples = jnp.asarray([0.5, 2., 2.1])
    num_live_points = jnp.asarray([1, 1, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # constraints are previous samples, with extra unfilled slots
    log_L_constraints = jnp.asarray([0., 0.5, 2., jnp.inf])
    log_L_samples = jnp.asarray([0.5, 2., 2.1, jnp.inf])
    num_live_points = jnp.asarray([1, 1, 1, 0])
    run_test(log_L_constraints, log_L_samples, num_live_points, num_samples=3, debug=True)

    # constraints are not ordered like samples (contour starts at non-minimum of constraints)
    log_L_constraints = jnp.asarray([0.45, 0., 2.])
    log_L_samples = jnp.asarray([0.5, 2., 2.1])
    num_live_points = jnp.asarray([2, 1, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # constraints are not ordered like samples (contour starts at non-minimum of constraints)
    log_L_constraints = jnp.asarray([0.45, 0., 2., jnp.inf, jnp.inf])
    log_L_samples = jnp.asarray([0.5, 2., 2.1, jnp.inf, jnp.inf])
    num_live_points = jnp.asarray([2, 1, 1, 0, 0])
    run_test(log_L_constraints, log_L_samples, num_live_points, num_samples=3, debug=True)

    log_L_constraints = jnp.asarray([0.45, 0.45, 0.5, 2., jnp.inf, jnp.inf])
    log_L_samples = jnp.asarray([0.5, 2., 2., 2.1, jnp.inf, jnp.inf])
    num_live_points = jnp.asarray([2, 2, 1, 1, 0, 0])
    run_test(log_L_constraints, log_L_samples, num_live_points, num_samples=4, debug=True)

    # constraints are not ordered like samples (contour starts at non-minimum of constraints)
    log_L_constraints = jnp.asarray([0., 0.45, 0., 2.])
    log_L_samples = jnp.asarray([0.5, 0.55, 2., 2.1])
    num_live_points = jnp.asarray([2, 2, 1, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # constraints are not ordered like samples (contour starts at non-minimum of constraints)
    log_L_constraints = jnp.asarray([0., 0., 0., 0.])
    log_L_samples = jnp.asarray([0.5, 0.5, 0.5, 0.5])
    num_live_points = jnp.asarray([4, 3, 2, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # constraints are not ordered like samples (contour starts at non-minimum of constraints)
    log_L_constraints = jnp.asarray([0., 0.45, 0.45, 0.])
    log_L_samples = jnp.asarray([0.5, 0.5, 0.5, 0.5])
    num_live_points = jnp.asarray([2, 1, 0, 0])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # All one common constraint, no plateau in samples
    log_L_constraints = jnp.asarray([0., 0., 0.])
    log_L_samples = jnp.asarray([0.5, 2., 2.1])
    num_live_points = jnp.asarray([3, 2, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # All one common constraint, no plateau in samples
    log_L_constraints = jnp.asarray([0., 0., 0., jnp.inf])
    log_L_samples = jnp.asarray([0.5, 2., 2.1, jnp.inf])
    num_live_points = jnp.asarray([3, 2, 1, 0])
    run_test(log_L_constraints, log_L_samples, num_live_points, num_samples=3, debug=True)

    # Growing
    log_L_constraints = jnp.asarray([0., 0.5, 0.5])
    log_L_samples = jnp.asarray([0.5, 2., 2.1])
    num_live_points = jnp.asarray([1, 2, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # Growing with plateau
    log_L_constraints = jnp.asarray([0., 0.5, 0.5, 2.])
    log_L_samples = jnp.asarray([0.5, 2., 2.1, 2.1])
    num_live_points = jnp.asarray([1, 2, 2, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # All one common constraint, a plateau in samples
    log_L_constraints = jnp.asarray([0., 0., 0., 0.])
    log_L_samples = jnp.asarray([0.5, 2., 2., 2.1])
    num_live_points = jnp.asarray([4, 3, 2, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # # one plateau -- broken assumption
    # log_L_constraints = jnp.asarray([0., 0., 0.])
    # log_L_samples = jnp.asarray([0., 0., 0.])
    # num_live_points = jnp.asarray([0, 0, 0])
    # run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # one plateau
    log_L_constraints = jnp.asarray([0., 0., 0.])
    log_L_samples = jnp.asarray([0.1, 0.1, 0.1])
    num_live_points = jnp.asarray([3, 2, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # constraints are not other samples
    # 0. (1) 0.5 (1) 1.5 (1) 2.5
    log_L_constraints = jnp.asarray([0., 0.25, 0.75])
    log_L_samples = jnp.asarray([0.5, 1.5, 2.5])
    num_live_points = jnp.asarray([1, 1, 1])
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # random generated example, no plateaus
    log_L_constraints = random.uniform(random.PRNGKey(42), shape=(100,))
    log_L_samples = random.uniform(random.PRNGKey(43), shape=(100,), minval=log_L_constraints)
    num_live_points = _sure_compute_num_live_points_from_unit_threads(log_L_constraints, log_L_samples)
    run_test(log_L_constraints, log_L_samples, num_live_points)

    # random generated example, no plateaus
    log_L_constraints = jnp.concatenate([jnp.zeros(5), 0.5 * jnp.arange(5) / 5 + 0.5])
    log_L_samples = random.uniform(random.PRNGKey(43), shape=(10,), minval=log_L_constraints)
    num_live_points = _sure_compute_num_live_points_from_unit_threads(log_L_constraints, log_L_samples)
    run_test(log_L_constraints, log_L_samples, num_live_points, debug=True)

    # print(num_live_points)
    # import pylab as plt
    # plt.plot(num_live_points)
    # plt.show()


def test_standard_nested_sampling():
    n = 0.5

    def log_likelihood(x):
        return 1. - x ** n

    def sample(key, log_L):
        # sample from 1 - x**n > log_L
        upper = (1. - log_L) ** (1. / n)
        return (1. - random.uniform(key)) * upper

    num_live_points = 10
    key = random.PRNGKey(42)
    num_samples = 100
    log_L_constraints = []
    log_L_samples = []
    U_samples = []

    key, sample_key = random.split(key, 2)

    U_samples_live = random.uniform(sample_key, (num_live_points,))
    log_L_samples_live = vmap(log_likelihood)(U_samples_live)
    log_L_constraints_live = jnp.asarray([0.] * num_live_points)

    while len(log_L_samples) < num_samples:
        idx_min = jnp.argmin(log_L_samples_live)
        log_L_samples.append(log_L_samples_live[idx_min])
        log_L_constraints.append(log_L_constraints_live[idx_min])
        U_samples.append(U_samples_live[idx_min])

        key, sample_key = random.split(key, 2)
        U_sample = sample(sample_key, log_L_samples[-1])
        log_L_sample = log_likelihood(U_sample)

        assert log_L_sample > log_L_samples[-1], "Numerical precision problem with unittest, don't shrink so much."

        log_L_samples_live = replace_index(log_L_samples_live, log_L_sample, idx_min)
        U_samples_live = replace_index(U_samples_live, U_sample, idx_min)
        log_L_constraints_live = replace_index(log_L_constraints_live, log_L_samples[-1], idx_min)

    for idx_min in jnp.argsort(log_L_samples_live):
        log_L_samples.append(log_L_samples_live[idx_min])
        log_L_constraints.append(log_L_constraints_live[idx_min])
        U_samples.append(U_samples_live[idx_min])

    n = compute_num_live_points_from_unit_threads(jnp.asarray(log_L_constraints), jnp.asarray(log_L_samples))
    for ni, log_L_constraint, log_L_sample in zip(n, jnp.asarray(log_L_constraints), jnp.asarray(log_L_samples)):
        # should be all `num_live_points` except last bit which decreases to 1
        print(ni, log_L_constraint, log_L_sample)

    assert jnp.allclose(n[:num_samples], num_live_points)

    n_check = _sure_compute_num_live_points_from_unit_threads(jnp.asarray(log_L_constraints),
                                                              jnp.asarray(log_L_samples))
    assert jnp.allclose(n, n_check)


def infimum_constraint(log_L_constraints, log_L_samples, sort_idx=None, return_contours: bool = False):
    """
    For a single sample find `i` such that log_L_contours[i] is the greatest strict lower bound of log_L_sample.
    E.g.

    Args:
        log_L_constraints:
        log_L_samples:
        sort_idx:
        return_contours: if true also return the value of the contour at `i`, i.e. the constraint.
    """
    if sort_idx is not None:
        log_L_constraints = log_L_constraints[sort_idx]
        log_L_samples = log_L_samples[sort_idx]
    # mask the non-samples, already done since they should be inf.
    # if num_samples is None:
    #       num_samples = log_L_samples.size
    # empty_mask = jnp.arange(log_L_samples.size) >= num_samples
    # log_L_constraints = jnp.where(empty_mask, jnp.inf, log_L_constraints)
    # log_L_samples = jnp.where(empty_mask, jnp.inf, log_L_samples)
    log_L_contours = jnp.concatenate([log_L_constraints[0:1], log_L_samples[:-1]], axis=0)
    contour_idx = jnp.searchsorted(log_L_contours, log_L_samples, side='left') - 1
    if return_contours:
        # todo: consider clamping to (0, n-1) and avoid the where op
        constraints = jnp.where(contour_idx < 0,
                                -jnp.inf,
                                log_L_contours[contour_idx])
        return contour_idx, constraints
    return contour_idx


def test_compute_remaining_evidence():
    # [a,b,-inf], 2 -> [a+b, b, -inf]
    log_dZ_mean = jnp.asarray([0., 1., -jnp.inf])
    sample_idx = 2
    expect = jnp.asarray([jnp.logaddexp(0, 1), 1, -jnp.inf])
    assert jnp.allclose(compute_remaining_evidence(sample_idx, log_dZ_mean), expect)

    # [-inf, -inf,-inf], 0 -> [-inf, -inf, -inf]
    log_dZ_mean = jnp.asarray([-jnp.inf, -jnp.inf - jnp.inf])
    sample_idx = 0
    expect = jnp.asarray([-jnp.inf, -jnp.inf - jnp.inf])
    assert jnp.allclose(compute_remaining_evidence(sample_idx, log_dZ_mean), expect)


def _create_test_reservoir(key: PRNGKey, num_live_points: int) -> Reservoir:
    n = 2

    # Prior is uniform in U[0,1]
    # Likelihood is 1 - x**n
    # Z = 1 - 1/n+1

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return (LogSpace(0.) - LogSpace(jnp.log(x)) ** n).log_abs_val

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    ns = ApproximateNestedSampler(
        model=model,
        num_live_points=num_live_points,
        num_parallel_samplers=1,
        max_samples=5,
        sampler_chain=[
            BadUniformSampler(mis_fraction=0., model=model)
        ]
    )

    termination_reason, state = ns(key, term_cond=TerminationCondition(live_evidence_frac=1e-4))
    return tree_map(lambda x: x[:state.sample_collection.sample_idx], state.sample_collection.reservoir)


def _exact_reservoir_num_live_points(reservoir: Reservoir) -> jnp.ndarray:
    num_live_points, _ = perfect_live_point_computation(reservoir.log_L_constraint, log_L_samples=reservoir.log_L)
    return num_live_points


def test_combine_reservoirs():
    reservoir_1 = _create_test_reservoir(random.PRNGKey(42), num_live_points=2)
    reservoir_2 = _create_test_reservoir(random.PRNGKey(43), num_live_points=2)
    print(reservoir_1)

    combined_reservoirs, num_live_points = combine_reservoirs(reservoir_1, reservoir_2)

    # print(combined_reservoirs)
    print(num_live_points)

    def make_contours(log_L_constraints, log_L_samples):
        sort_idx = jnp.lexsort((log_L_constraints, log_L_samples))
        log_L_samples = log_L_samples[sort_idx]
        log_L_constraints = log_L_constraints[sort_idx]
        log_L_contour = log_L_constraints[0]
        search_contours = jnp.concatenate([log_L_contour[None], log_L_samples], axis=0)

        contour_map_idx = jnp.searchsorted(search_contours, log_L_samples, side='left') - 1
        log_L_contours = search_contours[contour_map_idx]
        return log_L_contours

    print(compute_num_live_points_from_unit_threads(
        make_contours(combined_reservoirs.log_L_constraint, combined_reservoirs.log_L),
        combined_reservoirs.log_L,
        sorted_collection=True))



import logging

import jax.numpy as jnp
import numpy as np
import pylab as plt
from jax import random
from matplotlib.animation import FuncAnimation
from mpl_toolkits.axes_grid1 import make_axes_locatable

from jaxns.utils import resample
from jaxns.internals.log_semiring import cumulative_logsumexp
from jaxns.internals.shapes import tuple_prod
from jaxns.types import NestedSamplerResults, int_type

logger = logging.getLogger('jaxns')

__all__ = ['plot_diagnostics',
           'plot_cornerplot']


def plot_diagnostics(results: NestedSamplerResults, save_name=None):
    """
    Plot diagnostics of the nested sampling run.

    Args:
        results: NestedSamplerResult
        save_name: file to save figure to.
    """
    fig, axs = plt.subplots(5, 1, sharex=True, figsize=(8, 12))
    log_X = results.log_X_mean
    axs[0].plot(-log_X, results.num_live_points_per_sample)
    axs[0].set_ylabel(r'$n_{\rm live}(U)$')
    # detect if too small log likelihood
    log_likelihood = results.log_L_samples
    max_log_likelihood = jnp.max(log_likelihood)
    likelihood = jnp.exp(log_likelihood - max_log_likelihood)
    axs[1].plot(-log_X, likelihood)
    axs[1].hlines(1., jnp.min(-log_X),
                  jnp.max(-log_X), colors='black', ls='dashed',
                  label=r"$\log L_{{\rm max}}={:.1f}$".format(max_log_likelihood))
    axs[1].set_ylabel(r'$L(U)/L_{\rm max}$')
    axs[1].legend()
    axs[2].plot(-log_X, jnp.exp(results.log_dp_mean))
    # axs[2].vlines(-results.H_mean, 0., jnp.exp(jnp.max(results.log_dp_mean)), colors='black', ls='dashed',
    #               label='-logX=-H={:.1f}'.format(-results.H_mean))
    axs[2].set_ylabel(r'$Z^{-1}L(U) dX$')
    axs[2].legend()
    log_cum_evidence = cumulative_logsumexp(results.log_dp_mean)
    cum_evidence = jnp.exp(log_cum_evidence)
    axs[3].plot(-log_X, cum_evidence)
    axs[3].hlines(1., jnp.min(-log_X),
                  jnp.max(-log_X), colors='black', ls='dashed',
                  label=r"$\log Z={:.1f}$".format(results.log_Z_mean))
    axs[3].set_ylabel(r'$Z(x > U)/Z$')
    axs[3].legend()
    axs[4].plot(-log_X, 1./results.num_likelihood_evaluations_per_sample)
    axs[4].hlines(jnp.exp(results.log_efficiency), jnp.min(-log_X),
                  jnp.max(-log_X), colors='black', ls='dashed',
                  label='avg. eff.={:.3f}'.format(jnp.exp(results.log_efficiency)))
    axs[4].set_ylabel("sampler efficiency")
    axs[4].set_ylim(0., 1.05)
    axs[4].legend()
    if save_name is not None:
        fig.savefig(save_name)
    plt.show()


def plot_cornerplot(results: NestedSamplerResults, vars=None, save_name=None):
    """
    Plots a cornerplot of the posterior samples.

    Args:
        results: NestedSamplerResult
        vars: list of variable names to plot, or None.
        save_name: file to save result to.
    """
    rkey0 = random.PRNGKey(123496)
    vars = _get_vars(results, vars)
    ndims = _get_ndims(results, vars)
    figsize = min(20, max(4, int(2 * ndims)))
    fig, axs = plt.subplots(ndims, ndims, figsize=(figsize, figsize))
    if ndims == 1:
        axs = [[axs]]
    nsamples = results.total_num_samples
    max_like_idx = jnp.argmax(results.log_L_samples)
    map_idx = jnp.argmax(results.log_posterior_density)
    log_p = results.log_dp_mean
    nbins = max(10, int(jnp.sqrt(results.ESS)) + 1)
    lims = {}
    dim = 0
    for key in vars:  # sorted(results.samples.keys()):
        n1 = tuple_prod(results.samples[key].shape[1:])
        for i in range(n1):
            samples1 = results.samples[key].reshape((nsamples, -1))[:, i]
            if jnp.std(samples1) == 0.:
                dim += 1
                continue
            weights = jnp.where(jnp.isfinite(samples1), jnp.exp(log_p), 0.)
            log_weights = jnp.where(jnp.isfinite(samples1), log_p, -jnp.inf)
            samples1 = jnp.where(jnp.isfinite(samples1), samples1, 0.)
            # kde1 = gaussian_kde(samples1, weights=weights, bw_method='silverman')
            # samples1_resampled = kde1.resample(size=int(results.ESS))
            rkey0, rkey = random.split(rkey0, 2)
            samples1_resampled = resample(rkey, samples1, log_weights, S=max(10, int(results.ESS)), replace=True)
            samples1_max_like = samples1[max_like_idx]
            samples1_map_point = samples1[map_idx]
            binsx = jnp.linspace(*jnp.percentile(samples1_resampled, jnp.asarray([0, 100])), 2 * nbins)
            dim2 = 0
            for key2 in vars:  # sorted(results.samples.keys()):
                n2 = tuple_prod(results.samples[key2].shape[1:])
                for i2 in range(n2):
                    ax = axs[dim][dim2]
                    if dim2 > dim:
                        dim2 += 1
                        ax.set_xticks([])
                        ax.set_xticklabels([])
                        ax.set_yticks([])
                        ax.set_yticklabels([])
                        ax.remove()
                        continue
                    if dim < ndims - 1:
                        ax.set_xticks([])
                        ax.set_xticklabels([])
                    if dim2 > 0:
                        ax.set_yticks([])
                        ax.set_yticklabels([])
                    if n2 > 1:
                        title2 = "{}[{}]".format(key2, i2)
                    else:
                        title2 = "{}".format(key2)
                    if n1 > 1:
                        title1 = "{}[{}]".format(key, i)
                    else:
                        title1 = "{}".format(key)
                    # ax.set_title('{} {}'.format(title1, title2))
                    if dim == dim2:
                        # ax.plot(binsx, kde1(binsx))
                        ax.hist(np.asarray(samples1_resampled), bins='auto', fc='None', edgecolor='black', density=True)
                        ax.axvline(samples1_max_like, color='green')
                        sample_mean = jnp.average(samples1, weights=weights)
                        sample_std = jnp.sqrt(jnp.average((samples1 - sample_mean) ** 2, weights=weights))
                        ax.set_title(
                            r"${:.2f}_{{{:.2f}}}^{{{:.2f}}}$".format(
                                *jnp.percentile(samples1_resampled, jnp.asarray([50, 5, 95]))) + \
                            "\n" + r"${:.2f}\pm{:.2f}$".format(sample_mean, sample_std) + \
                            "\n" + r"MAP ${:.2f}$ | ML ${:.2f}$".format(samples1_map_point, samples1_max_like))
                        # ax.set_title(r"{}: ${:.2f}\pm{:.2f}$".format(title1, sample_mean, sample_std))
                        # ax.text(0., 1., r"${:.2f}_{{{:.2f}}}^{{{:.2f}}}$".format(*jnp.percentile(samples1_resampled, [50, 5, 95])),
                        #      verticalalignment = 'top', horizontalalignment='left', transform = ax.transAxes,
                        #         bbox=dict(facecolor='grey', alpha=0.5))

                        ax.axvline(sample_mean, linestyle='dashed', color='red')
                        ax.axvline(sample_mean + sample_std,
                                   linestyle='dotted', color='red')
                        ax.axvline(sample_mean - sample_std,
                                   linestyle='dotted', color='red')
                        ax.set_xlim(binsx.min(), binsx.max())
                        ax.set_yticks([])
                        ax.set_yticklabels([])
                        lims[dim] = ax.get_xlim()
                    else:
                        samples2 = results.samples[key2].reshape((nsamples, -1))[:, i2]
                        if jnp.std(samples2) == 0.:
                            dim2 += 1
                            continue
                        weights = jnp.where(jnp.isfinite(samples2), jnp.exp(log_p), 0.)
                        log_weights = jnp.where(jnp.isfinite(samples2), log_p, -jnp.inf)
                        samples2 = jnp.where(jnp.isfinite(samples2), samples2, 0.)
                        # kde2 = gaussian_kde(jnp.stack([samples1, samples2], axis=0),
                        #                     weights=weights,
                        #                     bw_method='silverman')
                        # samples2_resampled = kde2.resample(size=int(results.ESS))
                        rkey0, rkey = random.split(rkey0, 2)
                        samples2_resampled = resample(rkey, jnp.stack([samples1, samples2], axis=-1), log_weights,
                                                      S=max(10, int(results.ESS)), replace=True)
                        # norm = plt.Normalize(log_weights.min(), log_weights.max())
                        # color = jnp.atleast_2d(plt.cm.jet(norm(log_weights)))
                        ax.hist2d(samples2_resampled[:, 1], samples2_resampled[:, 0], bins=(nbins, nbins), density=True,
                                  cmap=plt.cm.bone_r)
                        # ax.scatter(samples2_resampled[:, 1], samples2_resampled[:, 0], marker='+', c='black', alpha=0.5)
                        # binsy = jnp.linspace(*jnp.percentile(samples2_resampled[:, 1], [0, 100]), 2 * nbins)
                        # U, Y = jnp.meshgrid(binsx, binsy, indexing='ij')
                        # ax.contour(kde2(jnp.stack([U.flatten(), Y.flatten()], axis=0)).reshape((2 * nbins, 2 * nbins)),
                        #            extent=(binsy.min(), binsy.max(),
                        #                    binsx.min(), binsx.max()),
                        #            origin='lower')
                    if dim == ndims - 1:
                        ax.set_xlabel("{}".format(title2))
                    if dim2 == 0:
                        ax.set_ylabel("{}".format(title1))

                    dim2 += 1
            dim += 1
    for dim in range(ndims):
        for dim2 in range(ndims):
            if dim == dim2:
                continue
            ax = axs[dim][dim2] if ndims > 1 else axs[0]
            if dim in lims.keys():
                ax.set_ylim(lims[dim])
            if dim2 in lims.keys():
                ax.set_xlim(lims[dim2])
    if save_name is not None:
        fig.savefig(save_name)
    plt.show()


def _get_ndims(results, vars):
    ndims = int(sum([tuple_prod(v.shape[1:]) for k, v in results.samples.items() if (k in vars)]))
    return ndims


def _get_vars(results, vars):
    if vars is None:
        vars = [k for k, v in results.samples.items()]
    vars = [v for v in vars if v in results.samples.keys()]
    vars = sorted(vars)
    return vars


def plot_samples_development(results, vars=None, save_name=None):
    """
    Animate the live points in a corner plot, visualising how the algorithm proceeds.
    Caution, this can be very slow as it plots a frame per sample.

    Args:
        results: NestedSamplingResult
        vars: list of variable names to plot, or None
        save_name: '.mp4' file to save animation to.
    """
    if save_name is None:
        raise ValueError("In order to plot the animation we must save it.")
    vars = _get_vars(results, vars)
    ndims = _get_ndims(results, vars)
    figsize = min(20, max(4, int(2 * ndims)))
    fig, axs = plt.subplots(ndims, ndims, figsize=(figsize, figsize))
    if ndims == 1:
        axs = [[axs]]
    weights = jnp.exp(results.log_p_mean)
    max_samples = weights.size
    norm = plt.Normalize(weights.min(), weights.max())
    to_colour = lambda w: plt.cm.jet(norm(w))

    def _get_artists(artists, start, stop):
        lims = {}
        dim = 0
        for key in vars:  # sorted(results.samples.keys()):
            n1 = tuple_prod(results.samples[key].shape[1:])
            for i in range(n1):
                samples1 = results.samples[key].reshape((max_samples, -1))[:, i]
                samples1 = samples1[start:stop]
                dim2 = 0
                for key2 in vars:  # sorted(results.samples.keys()):
                    n2 = tuple_prod(results.samples[key2].shape[1:])
                    for i2 in range(n2):
                        ax = axs[dim][dim2]
                        if dim2 > dim:
                            dim2 += 1
                            ax.set_xticks([])
                            ax.set_xticklabels([])
                            ax.set_yticks([])
                            ax.set_yticklabels([])
                            continue
                        if n2 > 1:
                            title2 = "{}[{}]".format(key2, i2)
                        else:
                            title2 = "{}".format(key2)
                        if n1 > 1:
                            title1 = "{}[{}]".format(key, i)
                        else:
                            title1 = "{}".format(key)
                        # ax.set_title('{} {}'.format(title1, title2))
                        if dim == dim2:
                            _, _, new_patches = ax.hist(samples1)
                            artists = artists + list(new_patches)
                            lims[dim] = ax.get_xlim()
                        else:
                            samples2 = results.samples[key2].reshape((max_samples, -1))[:, i2]
                            samples2 = samples2[start:stop]

                            sc = ax.scatter(samples2, samples1, marker='+', c=to_colour(weights[start:stop]), alpha=0.3)
                            artists.append(sc)
                        if dim == ndims - 1:
                            ax.set_xlabel("{}".format(title2))
                        if dim2 == 0:
                            ax.set_ylabel("{}".format(title1))

                        dim2 += 1
                dim += 1
        for dim in range(ndims):
            for dim2 in range(ndims):
                if dim == dim2:
                    continue
                ax = axs[dim][dim2] if ndims > 1 else axs[0]
                if dim in lims.keys():
                    ax.set_ylim(lims[dim])
                if dim2 in lims.keys():
                    ax.set_xlim(lims[dim2])
        return artists

    def init():
        start = 0
        stop = start + results.n_per_sample[start].astype(int_type)
        for i in range(ndims):
            for j in range(ndims):
                axs[i][j].clear()

        artists = []

        artists = _get_artists(artists, start, stop)
        return artists

    def update(start):
        stop = start + results.n_per_sample[start].astype(int_type)
        for i in range(ndims):
            for j in range(ndims):
                axs[i][j].clear()

        artists = []

        artists = _get_artists(artists, start, stop)
        return artists

    ani = FuncAnimation(fig, update, frames=jnp.arange(1, results.num_samples),
                        init_func=init, blit=True)

    ani.save(save_name, fps=results.n_per_sample[0] / 2.)


def add_colorbar_to_axes(ax, cmap, norm=None, vmin=None, vmax=None, label=None):
    """
    Add colorbar to axes easily.

    Args:
        ax: Axes
        cmap: str or cmap
        norm: Normalize or None
        vmin: lower limit of color if norm is None
        vmax: upper limit of color if norm is None
    """
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    if norm is None:
        norm = plt.Normalize(vmin=vmin, vmax=vmax)
    sm = plt.cm.ScalarMappable(norm, cmap=plt.cm.get_cmap(cmap))
    if label is None:
        ax.figure.colorbar(sm, cax=cax, orientation='vertical')
    else:
        ax.figure.colorbar(sm, cax=cax, orientation='vertical', label=label)


def corner_cornerplot(results: NestedSamplerResults):
    try:
        import corner
    except ImportError:
        logger.warning("You must run `pip install corner`")
        exit(0)
    try:
        import arviz as az
    except ImportError:
        logger.warning("You must run `pip install arviz`")
        exit(0)
    from jax import tree_map
    samples = resample(random.PRNGKey(42), results.samples, results.log_dp_mean, S=int(results.ESS))
    corner.corner(az.from_dict(posterior=tree_map(lambda x: x[None], samples)), )
    plt.show()



import inspect
import logging
from typing import TypeVar

from jax import tree_map, pmap, numpy as jnp
from jax.lax import dynamic_update_slice, dynamic_slice
from jax.lax import scan

from jaxns.types import int_type

logger = logging.getLogger(__name__)


def replace_index(operand, update, start_index):
    """
    Replaces an index or slice with an update.
    If update is too big to respect start_index then start_index is shifted, which will give non-intuitive results.
    """
    if len(operand.shape) != len(update.shape):
        update = update[None]
    start_index = jnp.asarray(start_index, int_type)
    start_indices = [start_index] + [jnp.asarray(0, start_index.dtype)] * (len(update.shape) - 1)
    return dynamic_update_slice(operand, update.astype(operand.dtype), start_indices)


def get_index(operand, start_index, length):
    return dynamic_slice(operand,
                         [start_index] + [jnp.asarray(0, int_type)] * (len(operand.shape) - 1),
                         (length,) + operand.shape[1:])


def prepare_func_args(f):
    """
    Takes a callable(a,b,...,z=Z) and prepares it into ``callable(**kwargs)``, such that only
    a,b,...,z are taken from ``**kwargs`` and the rest ignored.

    This allows ``f(**kwarg)`` to work even if ``f()`` is missing some keys from kwargs.

    Args:
        f: ``callable(a,b,...,z=Z)``

    Returns:
        ``callable(**kwargs)`` where ``**kwargs`` are the filtered for args of the original function.
    """
    if hasattr(f, "__old_name__"):
        raise ValueError(f"function {f.__old_name__} has already done prepare_func_args.")

    (args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations) = \
        inspect.getfullargspec(f)

    # TODO: this gets displayed each time we prepare a function. Using a cache would be cleaner for user.
    if varargs is not None:
        logger.warning(f"Function {f.__name__} has *varargs parameter ({varargs}), and is being dropped.")
    if varkw is not None:
        logger.warning(f"Function {f.__name__} has **varkw parameter ({varkw}), and is being dropped.")

    expected_keys = set(args + kwonlyargs)

    if defaults is None:  # no defaults
        num_defaults = 0
        defaults = ()
    else:
        num_defaults = len(defaults)

    num_non_default = len(args) - num_defaults

    def _f(**kwargs):
        # Replace any defaults with kwargs
        args_with_values = dict(zip(args[num_non_default:], defaults))
        if kwonlydefaults is not None:
            args_with_values.update(kwonlydefaults)
        args_with_values.update(kwargs)
        args_with_values = dict(filter(lambda item: item[0] in expected_keys, args_with_values.items()))
        for key in expected_keys:
            if key not in args_with_values:
                raise KeyError(f"{f.__name__} is missing argument {key}")

        return f(**args_with_values)

    _f.__doc__ = f.__doc__
    _f.__old_name__ = f.__name__
    return _f


F = TypeVar('F')


def chunked_pmap(f: F, chunksize, *, batch_size=None) -> F:
    def _f(*args, batch_size=batch_size, **kwargs):
        def queue(*args, **kwargs):
            """
            Distributes the computation in queues which are computed with scan.
            Args:
                *args:
            """

            def body(state, X):
                (args, kwargs) = X
                return state, f(*args, **kwargs)

            _, result = scan(body, (), (args, kwargs))
            return result

        if chunksize > 1:
            if batch_size is None:
                batch_size = args[0].shape[0] if len(args) > 0 else None
            assert batch_size is not None, "Couldn't get batch_size, please provide explicitly"
            remainder = batch_size % chunksize
            extra = (chunksize - remainder) % chunksize
            args = tree_map(lambda arg: _pad_extra(arg, chunksize), args)
            kwargs = tree_map(lambda arg: _pad_extra(arg, chunksize), kwargs)
            result = pmap(queue)(*args, **kwargs)
            result = tree_map(lambda arg: jnp.reshape(arg, (-1,) + arg.shape[2:]), result)
            if extra > 0:
                result = tree_map(lambda x: x[:-extra], result)
        else:
            result = queue(*args, **kwargs)
        return result

    _f.__doc__ = f.__doc__
    _f.__annotations__ = f.__annotations__
    return _f


def _pad_extra(arg, chunksize):
    N = arg.shape[0]
    remainder = N % chunksize
    if (remainder != 0) and (N > chunksize):
        # only pad if not a zero remainder
        extra = (chunksize - remainder) % chunksize
        arg = jnp.concatenate([arg] + [arg[0:1]] * extra, axis=0)
        N = N + extra
    else:
        extra = 0
    T = N // chunksize
    arg = jnp.reshape(arg, (chunksize, N // chunksize) + arg.shape[1:])
    return arg

def prepad(a, chunksize: int):
    return tree_map(lambda arg: _pad_extra(arg, chunksize), a)



from typing import Tuple, Union, List

from etils.array_types import FloatArray, IntArray
from jax import numpy as jnp, tree_map
from jax._src.lax.control_flow import while_loop
from jax._src.lax.slicing import dynamic_update_slice

from jaxns.internals.log_semiring import LogSpace
from jaxns.types import EvidenceCalculation, SampleCollection, SampleStatistics, float_type, int_type, Reservoir

__all__ = ['compute_evidence',
           'analyse_sample_collection',
           'compute_num_live_points_from_unit_threads']


def _init_evidence_calculation() -> EvidenceCalculation:
    """
    Returns an initial evidence calculation object.

    Returns:
        EvidenceCalculation
    """
    evidence_calculation = EvidenceCalculation(
        log_X_mean=jnp.asarray(0., float_type),
        log_X2_mean=jnp.asarray(0., float_type),
        log_Z_mean=jnp.asarray(-jnp.inf, float_type),
        log_ZX_mean=jnp.asarray(-jnp.inf, float_type),
        log_Z2_mean=jnp.asarray(-jnp.inf, float_type),
        log_dZ2_mean=jnp.asarray(-jnp.inf, float_type)
    )
    return evidence_calculation


def _update_evidence_calculation(num_live_points: FloatArray, log_L: FloatArray, next_log_L_contour: FloatArray,
                                 evidence_calculation: EvidenceCalculation) -> EvidenceCalculation:
    """
    Update an evidence calculation with the next sample from the shrinkage distribution.

    Args:
        num_live_points: number of live points at the current shrinkage step
        log_L: the log likelihood of the current contour
        next_log_L_contour: the log likelihood of the next contour after shrinkage
        evidence_calculation: the evidence calculation to update

    Returns:
        an updated evidence calculation
    """
    # num_live_points = num_live_points.astype(float_type)
    next_L = LogSpace(next_log_L_contour)
    L_contour = LogSpace(log_L)
    midL = LogSpace(jnp.log(0.5)) * (next_L + L_contour)
    X_mean = LogSpace(evidence_calculation.log_X_mean)
    X2_mean = LogSpace(evidence_calculation.log_X2_mean)
    Z_mean = LogSpace(evidence_calculation.log_Z_mean)
    ZX_mean = LogSpace(evidence_calculation.log_ZX_mean)
    Z2_mean = LogSpace(evidence_calculation.log_Z2_mean)
    dZ2_mean = LogSpace(evidence_calculation.log_dZ2_mean)

    # T_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.))
    # T_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points))
    T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
    # T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
    t_mean = LogSpace(- jnp.log(num_live_points + 1.))
    # T2_mean = LogSpace(jnp.log(num_live_points) - jnp.log( num_live_points + 2.))
    # T2_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 2./num_live_points))
    T2_mean = LogSpace(- jnp.logaddexp((0.), jnp.log(2.) - jnp.log(num_live_points)))
    # T2_mean = LogSpace(- jnp.logaddexp(jnp.log(2.), -jnp.log(num_live_points)))
    t2_mean = LogSpace(jnp.log(2.) - jnp.log(num_live_points + 1.) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points) - jnp.log(num_live_points + 2.))
    tT_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)) - jnp.log(num_live_points + 2.))

    next_X_mean = X_mean * T_mean
    next_X2_mean = X2_mean * T2_mean
    next_Z_mean = Z_mean + X_mean * t_mean * midL
    next_ZX_mean = ZX_mean * T_mean + X2_mean * tT_mean * midL
    next_Z2_mean = Z2_mean + LogSpace(jnp.log(2.)) * ZX_mean * t_mean * midL + (X2_mean * t2_mean * midL ** 2)
    next_dZ2_mean = dZ2_mean + (X2_mean * t2_mean * midL ** 2)

    next_evidence_calculation = evidence_calculation._replace(
        log_X_mean=next_X_mean.log_abs_val.astype(float_type),
        log_X2_mean=next_X2_mean.log_abs_val.astype(float_type),
        log_Z_mean=next_Z_mean.log_abs_val.astype(float_type),
        log_Z2_mean=next_Z2_mean.log_abs_val.astype(float_type),
        log_ZX_mean=next_ZX_mean.log_abs_val.astype(float_type),
        log_dZ2_mean=next_dZ2_mean.log_abs_val.astype(float_type)
    )
    # log_Z_mean, log_Z_var = linear_to_log_stats(next_evidence_calculation.log_Z_mean, log_f2_mean=next_evidence_calculation.log_Z2_mean)

    # next_evidence_calculation = tree_map(lambda old, new: jnp.where(jnp.isnan(new), old, new),
    #                                      evidence_calculation, next_evidence_calculation)

    return next_evidence_calculation


def compute_evidence(sample_collection: SampleCollection, num_live_points: FloatArray) \
        -> Tuple[EvidenceCalculation, SampleStatistics]:
    """
    Compute the evidence by traversing the sample collection.

    Args:
        sample_collection: the sorted sample collection to use to calculate the evidence
        num_live_points: the number of live points at each sample (see compute_num_live_points_from_unit_threads)

    Returns:
        evidence calculation and the sample statistics
    """
    num_samples = sample_collection.sample_idx

    CarryType = Tuple[EvidenceCalculation, IntArray, FloatArray, FloatArray, FloatArray]

    def thread_cond(body_state: CarryType):
        (evidence_calculation, idx, log_L_contour, log_dZ_mean, log_X_mean) = body_state
        return idx < num_samples

    def thread_body(body_state: CarryType) -> CarryType:
        (evidence_calculation, idx, log_L_contour, log_dZ_mean, log_X_mean) = body_state
        next_num_live_points = num_live_points[idx]
        next_log_L = sample_collection.reservoir.log_L[idx]
        next_log_L_contour = next_log_L
        # Get log_dZ_mean, and log_X_mean
        next_evidence_calculation = _update_evidence_calculation(
            num_live_points=next_num_live_points,
            log_L=log_L_contour,
            next_log_L_contour=next_log_L,
            evidence_calculation=evidence_calculation
        )

        next_dZ_mean = (LogSpace(next_evidence_calculation.log_Z_mean)
                        - LogSpace(evidence_calculation.log_Z_mean)).abs()

        next_log_dZ_mean = log_dZ_mean.at[idx].set(next_dZ_mean.log_abs_val)
        next_log_X_mean = log_X_mean.at[idx].set(next_evidence_calculation.log_X_mean)
        next_idx = idx + jnp.ones_like(idx)
        return (next_evidence_calculation, next_idx, next_log_L_contour, next_log_dZ_mean, next_log_X_mean)

    initial_evidence_calculation = _init_evidence_calculation()
    init_log_L_contour = sample_collection.reservoir.log_L_constraint[0]
    init_log_dZ_mean = jnp.full(num_live_points.shape, -jnp.inf, float_type)
    init_log_X_mean = jnp.full(num_live_points.shape, -jnp.inf, float_type)

    (final_evidence_calculation, final_idx, final_log_L_contour, final_log_dZ_mean, final_log_X_mean) = \
        while_loop(thread_cond,
                   thread_body,
                   (initial_evidence_calculation, jnp.asarray(0, int_type), init_log_L_contour,
                    init_log_dZ_mean, init_log_X_mean))

    sample_stats = SampleStatistics(
        num_live_points=num_live_points,
        log_X_mean=final_log_X_mean,
        log_dZ_mean=final_log_dZ_mean,
    )
    return final_evidence_calculation, sample_stats


def analyse_sample_collection(sample_collection: SampleCollection, sorted_collection: bool = True) \
        -> Tuple[EvidenceCalculation, SampleStatistics]:
    """
    Computes the evidence and statistics.

    Args:
        sample_collection: the sample collection
        sorted_collection: whether the sample collection is sorted

    Returns:
        evidence calculation and sample stats
    """
    # Sample collection has unsorted samples, and incorrect num_live_points.
    if sorted_collection:
        num_live_points = compute_num_live_points_from_unit_threads(
            log_L_constraints=sample_collection.reservoir.log_L_constraint,
            log_L_samples=sample_collection.reservoir.log_L,
            num_samples=sample_collection.sample_idx,
            sorted_collection=True
        )
    else:
        num_live_points, sort_idx = compute_num_live_points_from_unit_threads(
            log_L_constraints=sample_collection.reservoir.log_L_constraint,
            log_L_samples=sample_collection.reservoir.log_L,
            num_samples=sample_collection.sample_idx,
            sorted_collection=False
        )

        sample_collection = sample_collection._replace(
            reservoir=tree_map(lambda x: x[sort_idx], sample_collection.reservoir)
        )

    evidence_calculation, sample_stats = compute_evidence(
        sample_collection=sample_collection,
        num_live_points=num_live_points
    )

    return evidence_calculation, sample_stats


def perfect_live_point_computation_jax(log_L_constraints: jnp.ndarray, log_L_samples: jnp.ndarray,
                                       num_samples: Union[jnp.ndarray, None] = None):
    # log_L_constraints has shape [N]
    # log_L_samples has shape [N]
    sort_idx = jnp.lexsort((log_L_constraints, log_L_samples))
    log_L_samples = log_L_samples[sort_idx]
    log_L_constraints = log_L_constraints[sort_idx]
    log_L_contour = log_L_constraints[0]

    def outer_cond(state):
        (i, log_L_contour, log_L_samples, log_L_constraints, num_live_points) = state
        return i < log_L_samples.size

    def outer_loop_body(state):
        (i, log_L_contour, log_L_samples, log_L_constraints, num_live_points) = state
        log_L_dead = log_L_samples[i]

        def inner_cond(state):
            (i, log_L_samples, log_L_constraints, num_live_points) = state
            return jnp.bitwise_and(log_L_samples[i] == log_L_dead, i < log_L_constraints.size)

        def inner_loop_body(state):
            (i, log_L_samples, log_L_constraints, num_live_points) = state
            count = jnp.sum(jnp.bitwise_and(log_L_contour < log_L_samples, log_L_constraints <= log_L_contour))
            log_L_samples = log_L_samples.at[i].set(-jnp.inf)
            log_L_constraints = log_L_constraints.at[i].set(jnp.inf)
            num_live_points = num_live_points.at[i].set(count.astype(log_L_samples.dtype))
            return (i + 1, log_L_samples, log_L_constraints, num_live_points)

        (i, log_L_samples, log_L_constraints, num_live_points) = while_loop(
            inner_cond,
            inner_loop_body,
            (i, log_L_samples, log_L_constraints, num_live_points)
        )
        log_L_contour = log_L_dead
        return (i, log_L_contour, log_L_samples, log_L_constraints, num_live_points)

    (i, log_L_contour, log_L_samples, log_L_constraints, num_live_points) = while_loop(
        outer_cond,
        outer_loop_body,
        (jnp.asarray(0, int_type), log_L_contour, log_L_samples, log_L_constraints, jnp.zeros_like(log_L_samples))
    )
    if num_samples is not None:
        empty_mask = jnp.greater_equal(jnp.arange(log_L_samples.size), num_samples)
        num_live_points = jnp.where(empty_mask, 0, num_live_points)
    return num_live_points, sort_idx


def fast_triu_rowsum(a, b):
    """
    Computes a fast row sum of the upper triangular part of the outer product of a and b.

    Args:
        a: [N] array
        b: [N] array

    Returns:
        [N] of the row sum of upper trianguarl part of outer product of a and b.
    """
    b_cumsum_rev = jnp.cumsum(b[::-1])[::-1]
    return a * b_cumsum_rev


def fast_perfect_live_point_computation_jax(log_L_constraints: jnp.ndarray, log_L_samples: jnp.ndarray,
                                            num_samples: Union[jnp.ndarray, None] = None):
    # log_L_constraints has shape [N]
    # log_L_samples has shape [N]
    sort_idx = jnp.lexsort((log_L_constraints, log_L_samples))
    log_L_samples = log_L_samples[sort_idx]
    log_L_constraints = log_L_constraints[sort_idx]
    log_L_contour = log_L_constraints[0]
    search_contours = jnp.concatenate([log_L_contour[None], log_L_samples], axis=0)

    contour_map_idx = jnp.searchsorted(search_contours, log_L_samples, side='left') - 1
    log_L_contours = search_contours[contour_map_idx]
    diag_i = jnp.arange(log_L_samples.size)
    right_most_idx = jnp.searchsorted(jnp.sort(log_L_constraints), log_L_contours, side='right') - 1
    left_most_idx = jnp.maximum(diag_i, jnp.searchsorted(log_L_samples, log_L_contours, side='right') - 1)
    num_live_points = jnp.maximum(0, right_most_idx - left_most_idx + 1)

    if num_samples is not None:
        empty_mask = jnp.greater_equal(jnp.arange(log_L_samples.size), num_samples)
        num_live_points = jnp.where(empty_mask, jnp.asarray(0., log_L_samples.dtype), num_live_points)

    return num_live_points, sort_idx


def compute_num_live_points_from_unit_threads(log_L_constraints: FloatArray, log_L_samples: FloatArray,
                                              num_samples: IntArray = None, sorted_collection: bool = True) \
        -> Union[FloatArray, Tuple[FloatArray, IntArray]]:
    """
    Compute the number of live points of shrinkage distribution, from an arbitrary list of samples with
    corresponding sampling constraints.

    Args:
        log_L_constraints: [N] likelihood constraint that sample was uniformly sampled within
        log_L_samples: [N] likelihood of the sample
        sorted_collection: bool, whether the sample collection was already sorted.

    Returns:
        if sorted_collection is true:
            num_live_points for shrinkage distribution
        otherwise:
            num_live_points for shrinkage distribution, and sort indicies
    """
    num_live_points, sort_idx = fast_perfect_live_point_computation_jax(log_L_constraints=log_L_constraints,
                                                                        log_L_samples=log_L_samples,
                                                                        num_samples=num_samples)

    if not sorted_collection:
        return num_live_points, sort_idx

    return num_live_points


def compute_remaining_evidence(sample_idx, log_dZ_mean):
    # Z_remaining = dZ_mean.cumsum(reverse=True)

    def logsumexp_cumsum_body(_state):
        (log_abs_val, idx) = _state
        next_idx = idx - jnp.ones_like(idx)
        next_val = LogSpace(log_abs_val[idx]) + LogSpace(log_abs_val[next_idx])
        next_log_abs_val = dynamic_update_slice(log_abs_val, next_val.log_abs_val[None], [next_idx])
        return (next_log_abs_val, next_idx)

    # Calculate remaining evidence, doing only the minimal amount of work necessary.

    (log_Z_remaining, _) = while_loop(lambda _state: _state[1] > 0,
                                      logsumexp_cumsum_body,
                                      (log_dZ_mean,
                                       sample_idx - jnp.ones_like(sample_idx)))
    return log_Z_remaining


def combine_reservoirs(*reservoirs: Reservoir) -> Tuple[Reservoir, jnp.ndarray]:
    """
    Combine two reservoirs sorting, and updating the number of live points properly.

    Args:
        *reservoirs: Reservoirs to combine.

    Returns:
        a sorted, updated reservoir
    """
    if len(reservoirs) == 0:
        raise ValueError(f"Need at least one reservoir to combine.")
    if len(reservoirs) == 1:
        return reservoirs[0]

    def sort_reservoir(reservoir: Reservoir) -> Reservoir:
        sort_idx = jnp.lexsort((reservoir.log_L_constraint, reservoir.log_L))
        return tree_map(lambda x: x[sort_idx], reservoir)

    reservoirs: List[Reservoir] = list(map(sort_reservoir, reservoirs))

    # point_U -> concatenate and sort
    # log_L_constraint -> concatenate and sort
    # log_L -> concatenate and sort
    # num_likelihood_evaluations -> concatenate and sort
    # num_live_points -> concatenate, sort, and sum
    # num_slices -> concatenate and sort
    # iid -> concatenate and sort

    output_reservoir: Reservoir = tree_map(lambda *x: jnp.concatenate(x, axis=0), *reservoirs)
    sort_idx = jnp.lexsort((output_reservoir.log_L_constraint, output_reservoir.log_L))
    output_reservoir: Reservoir = tree_map(lambda x: x[sort_idx], output_reservoir)

    # combine each reservoir iteratively
    combined_num_live_points = jnp.zeros_like(output_reservoir.log_L)

    for r in reservoirs:
        num_live_points = compute_num_live_points_from_unit_threads(r.log_L_constraint, r.log_L,
                                                                    sorted_collection=True)
        print(num_live_points)
        j = 0
        for i, log_L in enumerate(r.log_L):
            while log_L > output_reservoir.log_L[j]:
                j += 1
            combined_num_live_points = combined_num_live_points.at[j].add(num_live_points[i])

    return output_reservoir, combined_num_live_points



from typing import NamedTuple

from etils.array_types import BoolArray, IntArray
from etils.array_types import PRNGKey, FloatArray
from jax import random, numpy as jnp
from jax.lax import while_loop

from jaxns.model import Model
from jaxns.static_nested_sampler import PreProcessType, RejectionSampler
from jaxns.types import NestedSamplerState, LivePoints, Sample, int_type

__all__ = ['UniformSampler']


class UniformSampler(RejectionSampler):
    def preprocess(self, state: NestedSamplerState) -> PreProcessType:
        return ()

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: PreProcessType) -> Sample:
        class CarryState(NamedTuple):
            done: BoolArray
            key: PRNGKey
            U: FloatArray
            log_L: FloatArray
            log_L_constraint: FloatArray
            num_likelihood_evals: IntArray

        def body(carry_state: CarryState):
            key, sample_key = random.split(carry_state.key, 2)
            log_L = self.model.forward(U=carry_state.U)
            num_likelihood_evals = carry_state.num_likelihood_evals + jnp.asarray(1, int_type)
            # backoff by one e-fold per attempt after efficiency threshold reached
            log_L_constraint = jnp.where(num_likelihood_evals > 1. / self.efficiency_threshold,
                                         carry_state.log_L_constraint - 0.1, carry_state.log_L_constraint)
            done = log_L > log_L_constraint
            U = jnp.where(done, carry_state.U, self.model.sample_U(key=sample_key))
            return CarryState(done=done, key=key, U=U, log_L=log_L, num_likelihood_evals=num_likelihood_evals,
                              log_L_constraint=log_L_constraint)

        key, sample_key = random.split(key, 2)
        init_carry_state = CarryState(done=jnp.asarray(False),
                                      key=key,
                                      U=self.model.sample_U(key=sample_key),
                                      log_L=log_L_constraint,
                                      log_L_constraint=log_L_constraint,
                                      num_likelihood_evals=jnp.asarray(0, int_type))

        carry_state = while_loop(lambda s: jnp.bitwise_not(s.done), body, init_carry_state)

        sample = Sample(point_U=carry_state.U,
                        log_L_constraint=carry_state.log_L_constraint,
                        log_L=carry_state.log_L,
                        num_likelihood_evaluations=carry_state.num_likelihood_evals,
                        num_slices=jnp.asarray(0, int_type),
                        iid=jnp.asarray(True, jnp.bool_))
        return sample


class BadUniformSampler(RejectionSampler):
    def __init__(self, mis_fraction: float, model: Model):
        super().__init__(model=model, efficiency_threshold=None)
        self.mis_fraction = mis_fraction

    def preprocess(self, state: NestedSamplerState) -> PreProcessType:
        return ()

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, live_points: LivePoints,
                   preprocess_data: PreProcessType) -> Sample:
        class CarryState(NamedTuple):
            done: BoolArray
            key: PRNGKey
            U: FloatArray
            log_L: FloatArray
            log_L_constraint: FloatArray
            num_likelihood_evals: IntArray

        def body(carry_state: CarryState):
            key, sample_key = random.split(carry_state.key, 2)
            log_L = self.model.forward(U=carry_state.U)
            num_likelihood_evals = carry_state.num_likelihood_evals + jnp.asarray(1, int_type)
            bad_log_L_constraint = carry_state.log_L_constraint + jnp.log(1. + self.mis_fraction)
            done = log_L > bad_log_L_constraint
            U = jnp.where(done, carry_state.U, self.model.sample_U(key=sample_key))
            return CarryState(done=done, key=key, U=U, log_L=log_L,
                              num_likelihood_evals=num_likelihood_evals,
                              log_L_constraint=carry_state.log_L_constraint)

        key, sample_key = random.split(key, 2)
        init_carry_state = CarryState(done=jnp.asarray(False),
                                      key=key,
                                      U=self.model.sample_U(key=sample_key),
                                      log_L=log_L_constraint,
                                      log_L_constraint=log_L_constraint,
                                      num_likelihood_evals=jnp.asarray(0, int_type))

        carry_state = while_loop(lambda s: jnp.bitwise_not(s.done), body, init_carry_state)

        sample = Sample(point_U=carry_state.U,
                        log_L_constraint=carry_state.log_L_constraint,
                        log_L=carry_state.log_L,
                        num_likelihood_evaluations=carry_state.num_likelihood_evals,
                        num_slices=jnp.asarray(0, int_type),
                        iid=jnp.asarray(self.mis_fraction == 0., jnp.bool_))
        return sample



from time import monotonic_ns

import pylab as plt
import tensorflow_probability.substrates.jax as tfp
from jax import random, numpy as jnp
from jax._src.lax.control_flow import scan
from jax._src.scipy.linalg import solve_triangular

from jaxns.internals.log_semiring import LogSpace
from jaxns.model import Model
from jaxns.nested_sampler import ApproximateNestedSampler, ExactNestedSampler
from jaxns.prior import PriorModelGen, Prior
from jaxns.static_uniform import BadUniformSampler
from jaxns.types import TerminationCondition, float_type
from jaxns.utils import sample_evidence, bruteforce_evidence

tfpd = tfp.distributions


class Timer:
    def __enter__(self):
        self.t0 = monotonic_ns()

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f"Time to execute: {(monotonic_ns() - self.t0) / 1e9} seconds.")


def test_approximate_nested_sampler():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=jnp.asarray([1., 1.])))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -jnp.sum(z ** 2)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    approx_ns = ApproximateNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                         max_samples=1000)
    # print(termination_reason)
    with Timer():
        termination_reason, state = approx_ns(random.PRNGKey(42),
                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = approx_ns.to_results(state, termination_reason)
        results.log_Z_mean.block_until_ready()
    approx_ns.summary(approx_ns.to_results(state, termination_reason))

    with Timer():
        termination_reason, state = approx_ns(random.PRNGKey(42),
                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = approx_ns.to_results(state, termination_reason)
        results.log_Z_mean.block_until_ready()


def test_exact_nested_sampler():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    model.sanity_check(random.PRNGKey(0), S=100)

    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1, max_samples=1000,
                                  patience=1)
    # print(termination_reason)
    # print(state)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        exact_ns.summary(results)
        exact_ns.plot_diagnostics(results)

    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)


def test_nested_sampling_basic():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
        return x

    def log_likelihood(x):
        return - jnp.sum(x ** 2)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                  max_samples=1000)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)

        # exact_ns.plot_diagnostics(results)
        exact_ns.summary(results)
        # exact_ns.plot_cornerplot(results)

    log_Z_true = bruteforce_evidence(model=model, S=200)

    log_Z_samples = sample_evidence(random.PRNGKey(42),
                                    results.num_live_points_per_sample,
                                    results.log_L_samples, S=1000)

    assert jnp.isclose(results.log_Z_mean, jnp.mean(log_Z_samples), atol=1e-3)
    assert jnp.isclose(results.log_Z_uncert, jnp.std(log_Z_samples), atol=1e-2)

    assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
    assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=1.75 * results.log_Z_uncert)


#
# def test_nested_sampling_basic_parallel():
#     def prior_model() -> PriorModelGen:
#         x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
#         return x
#
#     def log_likelihood(x):
#         return - jnp.sum(x ** 2)
#
#     model = Model(prior_model=prior_model,
#                   log_likelihood=log_likelihood)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=2,
#                                   max_samples=1000)
#
#     termination_reason, state = exact_ns(random.PRNGKey(42),
#                                          term_cond=TerminationCondition(live_evidence_frac=1e-4))
#     results = exact_ns.to_results(state, termination_reason)
#
#     exact_ns.plot_diagnostics(results)
#     exact_ns.summary(results)
#     exact_ns.plot_cornerplot(results)
#
#     log_Z_true = bruteforce_evidence(model=model, S=200)
#
#     log_Z_samples = sample_evidence(random.PRNGKey(42),
#                                                results.num_live_points_per_sample,
#                                                results.log_L_samples, S=1000)
#
#     assert jnp.isclose(results.log_Z_mean, jnp.mean(log_Z_samples), atol=1e-3)
#     assert jnp.isclose(results.log_Z_uncert, jnp.std(log_Z_samples), atol=1e-3)
#
#     assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
#     assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=1.75 * results.log_Z_uncert)


def test_nested_sampling_mvn_static():
    def log_normal(x, mean, cov):
        L = jnp.linalg.cholesky(cov)
        dx = x - mean
        dx = solve_triangular(L, dx, lower=True)
        return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) \
            - 0.5 * dx @ dx

    ndims = 8
    prior_mu = 15 * jnp.ones(ndims)
    prior_cov = jnp.diag(jnp.ones(ndims)) ** 2

    data_mu = jnp.zeros(ndims)
    data_cov = jnp.diag(jnp.ones(ndims)) ** 2
    data_cov = jnp.where(data_cov == 0., 0.99, data_cov)

    true_logZ = log_normal(data_mu, prior_mu, prior_cov + data_cov)
    # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
    post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
        prior_cov + data_cov) @ prior_mu

    print(f"True post mu:{post_mu}")
    print(f"True log Z: {true_logZ}")

    def prior_model() -> PriorModelGen:
        x = yield Prior(
            tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
            name='x')
        return x

    def log_likelihood(x):
        return log_normal(x, data_mu, data_cov)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    # model.sanity_check(random.PRNGKey(42), S=100)
    exact_ns = ExactNestedSampler(model=model, num_live_points=100, num_parallel_samplers=1,
                                  max_samples=40000, patience=20)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(41),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-5))
        results = exact_ns.to_results(state, termination_reason)
        exact_ns.summary(results)
    exact_ns.plot_diagnostics(results)
    actual_log_Z_mean = results.log_Z_mean
    expected_log_Z_mean = true_logZ
    tol = 1.75 * results.log_Z_uncert
    assert jnp.isclose(actual_log_Z_mean, expected_log_Z_mean, atol=tol)


# def test_multi_needs_fewer_slices():
#     def log_normal(x, mean, cov):
#         L = jnp.linalg.cholesky(cov)
#         dx = x - mean
#         dx = solve_triangular(L, dx, lower=True)
#         return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) - 0.5 * dx @ dx
#
#     ndims = 4
#     prior_mu = 2 * jnp.ones(ndims)
#     prior_cov = jnp.diag(jnp.ones(ndims)) ** 2
#
#     data_mu = jnp.zeros(ndims)
#     data_cov = jnp.diag(jnp.ones(ndims)) ** 2
#     data_cov = jnp.where(data_cov == 0., 0.95, data_cov)
#
#     true_logZ = log_normal(data_mu, prior_mu, prior_cov + data_cov)
#     # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
#     post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
#         prior_cov + data_cov) @ prior_mu
#
#     print(f"True post mu:{post_mu}")
#     print(f"True log Z: {true_logZ}")
#
#     def prior_model() -> PriorModelGen:
#         x = yield Prior(
#             tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
#             name='x')
#         return x
#
#     def log_likelihood(x):
#         return log_normal(x, data_mu, data_cov)
#
#     model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
#
#     model.sanity_check(random.PRNGKey(52), S=100)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
#                                   max_samples=1e4)
#     with Timer():
#         print("Using Multi Dimensional Slice Sampling")
#         termination_reason, state = exact_ns(random.PRNGKey(42),
#                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
#         results = exact_ns.to_results(state, termination_reason)
#         exact_ns.summary(results)
#         num_slices_multi = results.total_num_slices
#     assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)
#
#     slice_sampler = UniDimSliceSampler(model=model, midpoint_shrink=True)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
#                                   max_samples=1e4, refinement_slice_sampler=slice_sampler)
#     with Timer():
#         print("Using Uni Dimensional Slice Sampling")
#         termination_reason, state = exact_ns(random.PRNGKey(42),
#                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
#         results = exact_ns.to_results(state, termination_reason)
#         exact_ns.summary(results)
#         num_slices_uni = results.total_num_slices
#     assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)
#     # TODO(Joshuaalbert): broken test, probably means multidim slice sampling is not robust to structure.
#     # assert num_slices_multi < num_slices_uni


def test_shrinkage():
    n = 2

    # Prior is uniform in U[0,1]
    # Likelihood is 1 - x**n
    # Z = 1 - 1/n+1

    log_Z_true = jnp.log(1. - 1. / (n + 1))

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return jnp.log(1. - x ** n)

    def exact_X(L):
        return (1. - L) ** (1. / n)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                  max_samples=1000)

    termination_reason, state = exact_ns(random.PRNGKey(42),
                                         term_cond=TerminationCondition(live_evidence_frac=1e-4))
    results = exact_ns.to_results(state, termination_reason)

    # print(list(results.log_X_mean[:results.total_num_samples]))
    # print(list(results.log_L_samples[:results.total_num_samples]))
    # print(list(jnp.log(exact_X(jnp.exp(results.log_L_samples[:results.total_num_samples])))))
    diff = results.log_X_mean - jnp.log(exact_X(jnp.exp(results.log_L_samples)))
    diff = jnp.where(jnp.isfinite(diff), diff, jnp.nan)
    # print(jnp.nanstd(diff))
    assert jnp.nanstd(diff) < 0.26
    assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=results.log_Z_uncert * 1.75)


def compute_shrinkage_stats(num_live_points):
    def _update_stats(state, num_live_points):
        X_mean = LogSpace(state['log_X_mean'])
        X2_mean = LogSpace(state['log_X2_mean'])

        T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
        T2_mean = LogSpace(- jnp.logaddexp((0.), jnp.log(2.) - jnp.log(num_live_points)))

        next_X_mean = X_mean * T_mean
        next_X2_mean = X2_mean * T2_mean

        return dict(log_X_mean=next_X_mean.log_abs_val, log_X2_mean=next_X2_mean.log_abs_val), dict(
            log_X_mean=next_X_mean.log_abs_val, log_X2_mean=next_X2_mean.log_abs_val)

    state = dict(
        log_X_mean=jnp.asarray(0., float_type),
        log_X2_mean=jnp.asarray(0., float_type),
    )
    _, stats = scan(_update_stats,
                    state,
                    num_live_points)
    log_X_uncert = (LogSpace(stats['log_X2_mean']) - LogSpace(stats['log_X_mean']).square()).sqrt().log_abs_val
    return stats['log_X_mean'], log_X_uncert


def test_adaptive_refinement():
    n = 2

    # Prior is uniform in U[0,1]
    # Likelihood is 1 - x**n
    # Z = 1 - 1/n+1

    log_Z_true = jnp.log(1. - 1. / (n + 1))
    print(f"True log(Z): {log_Z_true}")

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return (LogSpace(0.) - LogSpace(n * jnp.log(x))).log_abs_val

    def exact_X(L):
        return (1. - L) ** (1. / n)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    ns = ApproximateNestedSampler(
        model=model,
        num_live_points=50,
        num_parallel_samplers=1,
        max_samples=1e4,
        sampler_chain=[
            BadUniformSampler(mis_fraction=0., model=model)
        ]
    )

    termination_reason, state = ns(random.PRNGKey(42),
                                   term_cond=TerminationCondition(live_evidence_frac=1e-4))
    results = ns.to_results(state, termination_reason)
    ns.summary(results)
    ns.plot_diagnostics(results)

    # ensure there is no bug in control flow. X should be same to controlled evaluation
    log_X_mean, log_X_std = compute_shrinkage_stats(results.num_live_points_per_sample)
    assert jnp.allclose(results.log_X_mean, log_X_mean)

    # ensure the deviation from the exact shrinkage is correct
    X_exact = exact_X(jnp.exp(results.log_L_samples))

    rel_diff = jnp.abs(jnp.exp(log_X_mean) - X_exact) / jnp.exp(log_X_std)

    # ensure log_Z is close to truth
    assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=results.log_Z_uncert * 1.75)

    print("Relative shrinkage errors", jnp.percentile(rel_diff, jnp.asarray([50, 75, 90, 95])))
    assert jnp.all(jnp.percentile(rel_diff, jnp.asarray([50, 75, 90, 95])) < jnp.asarray([0.9, 1.1, 1.4, 1.5]))


def test_nested_sampling_plateau():
    def log_likelihood(x):
        return 0.

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(
        model=model,
        num_live_points=50,
        max_samples=1000
    )

    termination_reason, state = exact_ns(
        random.PRNGKey(42),
        term_cond=TerminationCondition(live_evidence_frac=1e-4)
    )

    results = exact_ns.to_results(state, termination_reason)
    exact_ns.summary(results)
    exact_ns.plot_diagnostics(results)

    log_Z_samples = sample_evidence(random.PRNGKey(42),
                                    results.num_live_points_per_sample,
                                    results.log_L_samples, S=1000)
    plt.hist(log_Z_samples, bins='auto')
    plt.show()

    true_log_Z_uncert = jnp.std(log_Z_samples)
    print(f"true logZ uncert: {true_log_Z_uncert}")

    assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
    assert jnp.isclose(results.log_Z_mean, 0., atol=1.75 * true_log_Z_uncert)



import pylab as plt
from jax import numpy as jnp, random, jit
from sklearn.datasets import make_blobs
from sklearn.metrics import adjusted_rand_score

from jaxns.likelihood_samplers.em_gmm import em_gmm


def test_em_gmm():
    data = jnp.array([[1.0, 1.0], [1.1, 1.1], [1.2, 1.2], [4.0, 4.0], [4.1, 4.1], [4.2, 4.2]])
    n_components = 2
    key = random.PRNGKey(42)

    cluster_id, (means, covariances, log_weights), _ = em_gmm(key, data, n_components)
    assert jnp.all(cluster_id == jnp.asarray([1, 1, 1, 0, 0, 0]))


def test_blob_decomp():
    key = random.PRNGKey(42)

    d = 2
    n_data = 500
    jit_em_gmm = jit(em_gmm, static_argnames=['n_components', 'n_iters', 'tol'])

    for n_components in [2, 3]:
        X, y_true = make_blobs(n_samples=n_data, centers=n_components, n_features=d, cluster_std=1, random_state=42)

        cluster_id, (means, covariances, log_weights), total_iters = jit_em_gmm(key, X, n_components=n_components,
                                                                                n_iters=100)
        print("Used", total_iters)
        plt.scatter(X[:, 0], X[:, 1], c=cluster_id, cmap='jet')
        # plt.show()
        plt.close('all')
        accuracy = adjusted_rand_score(y_true, cluster_id)
        print(accuracy)
        assert accuracy == 1.



from jax import numpy as jnp, vmap

from jaxns.internals.log_semiring import LogSpace
from jaxns.types import float_type


def normal_to_lognormal(mu, std):
    """
    Convert normal parameters to log-normal parameters.
    Args:
        mu:
        var:

    Returns:

    """
    var = std ** 2
    ln_mu = 2. * jnp.log(mu) - 0.5 * jnp.log(var)
    ln_var = jnp.log(var) - 2. * jnp.log(mu)
    return ln_mu, jnp.sqrt(ln_var)


def density_estimation(xstar, x, alpha=1. / 3., order=1):
    assert len(x.shape) == 2

    N = x.shape[0]
    m = int(pow(N, 1. - alpha))
    s = N // m
    N = m * s
    x = x[:N]

    def single_density(xstar):
        dist = jnp.linalg.norm(x - xstar, ord=order, axis=-1)  # N
        dist = jnp.reshape(dist, (s, m))  # s,m
        min_dist = jnp.min(dist, axis=0)  # m
        avg_dist = jnp.mean(min_dist)  # scalar
        return 0.5 / ((1. + s) * avg_dist)

    return vmap(single_density)(xstar)


def linear_to_log_stats(log_f_mean, *, log_f2_mean=None, log_f_var=None):
    """
    Converts normal to log-normal stats.
    Args:
        log_f_mean: log(E(f))
        log_f2_mean: log(E(f**2))
        log_f_var: log(Var(f))
    Returns:
        E(log(f))
        Var(log(f))
    """
    f_mean = LogSpace(log_f_mean)
    if log_f_var is not None:
        f_var = LogSpace(log_f_var)
        f2_mean = f_var + f_mean.square()
    else:
        f2_mean = LogSpace(log_f2_mean)
    mu = f_mean.square() / f2_mean.sqrt()
    sigma2 = f2_mean / f_mean.square()
    return mu.log_abs_val, jnp.maximum(sigma2.log_abs_val, jnp.finfo(float_type).eps)


def effective_sample_size(log_Z_mean, log_dZ2_mean):
    """
    Computes Kish's ESS = [sum dZ]^2 / [sum dZ^2]

    :param log_Z_mean:
    :param log_dZ2_mean:
    :return:
    """
    ess = LogSpace(log_Z_mean).square() / LogSpace(log_dZ2_mean)
    return ess.value



import logging
from abc import abstractmethod, ABC
from typing import Tuple, Generator, Callable, Optional, Union, List

import jax.numpy as jnp
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from etils.array_types import FloatArray, IntArray, BoolArray

from jaxns.internals.shapes import tuple_prod
from jaxns.types import float_type, LikelihoodType, LikelihoodInputType, UType, XType

logger = logging.getLogger('jaxns')
tfpd = tfp.distributions

__all__ = [
    "Prior",
    "PriorModelGen"
]


class InvalidDistribution(Exception):
    def __init__(self, dist: Optional[tfpd.Distribution] = None):
        super(InvalidDistribution, self).__init__(
            f'Distribution {dist} is missing a quantile. '
            f'Try checking if your desired prior exists in `jaxns.special_priors`.')


class InvalidPriorName(Exception):
    def __init__(self, name: Optional[str] = None):
        super(InvalidPriorName, self).__init__(f'Prior name {name} already taken by another prior.')


def distribution_chain(dist: tfpd.Distribution) -> List[Union[tfpd.TransformedDistribution,
tfpd.Sample,
tfpd.Distribution]]:
    chain = []
    while True:
        chain.append(dist)
        if isinstance(dist, tfpd.TransformedDistribution):
            dist = dist.distribution
            continue
        break
    return chain[::-1]


class AbstractDistribution(ABC):
    @abstractmethod
    def _dtype(self):
        ...

    @abstractmethod
    def _base_shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _forward(self, U):
        ...

    @abstractmethod
    def _inverse(self, X):
        ...

    @abstractmethod
    def _log_prob(self, X):
        ...

    @property
    def dtype(self):
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        return self._base_shape()

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape()

    def forward(self, U):
        return self._forward(U)

    def inverse(self, X):
        return self._inverse(X)

    def log_prob(self, X):
        return self._log_prob(X)


class Distribution(AbstractDistribution):
    def __init__(self, dist: tfpd.Distribution):
        self.dist_chain = distribution_chain(dist)
        check_dist = self.dist_chain[0]
        if isinstance(self.dist_chain[0], tfpd.Sample):
            check_dist = self.dist_chain[0].distribution
        if '_quantile' not in check_dist.__class__.__dict__:
            # TODO(Joshuaalbert): we could numerically approximate it. This requires knowing the support of dist.
            # Repartitioning the prior also requires knowing the support and choosing a replacement, which is not
            # always easy from stats. E.g. StudentT variance doesn't exist but a numerial quantile can be made.
            raise InvalidDistribution(dist=dist)

    def __repr__(self):
        return " -> ".join(map(repr, self.dist_chain))

    def _dtype(self):
        return self.dist_chain[-1].dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return tuple(self.dist_chain[0].batch_shape_tensor()) + tuple(self.dist_chain[0].event_shape_tensor())

    def _shape(self) -> Tuple[int, ...]:
        return tuple(self.dist_chain[-1].batch_shape_tensor()) + tuple(self.dist_chain[-1].event_shape_tensor())

    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        dist = self.dist_chain[0]
        if isinstance(dist, tfpd.Sample):
            dist = dist.distribution
        X = dist.quantile(U)
        for dist in self.dist_chain[1:]:
            X = dist.bijector.forward(X)
        return X

    def _inverse(self, X) -> FloatArray:
        for dist in reversed(self.dist_chain[1:]):
            X = dist.bijector.inverse(X)
        dist = self.dist_chain[0]
        if isinstance(dist, tfpd.Sample):
            dist = dist.distribution
        X = dist.cdf(X)
        return X

    def _log_prob(self, X):
        return self.dist_chain[-1].log_prob(X)


class AbstractPrior(ABC):
    def __init__(self, name: Optional[str] = None):
        self.name = name

    def __repr__(self):
        return f"{self.name if self.name is not None else '*'}\t{self.base_shape} -> {self.shape} {self.dtype}"

    @abstractmethod
    def _dtype(self):
        ...

    @abstractmethod
    def _base_shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _shape(self) -> Tuple[int, ...]:
        ...

    @abstractmethod
    def _forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        ...

    @abstractmethod
    def _inverse(self, X) -> FloatArray:
        ...

    @abstractmethod
    def _log_prob(self, X) -> FloatArray:
        ...

    @property
    def dtype(self):
        return self._dtype()

    @property
    def base_shape(self) -> Tuple[int, ...]:
        return self._base_shape()

    @property
    def base_ndims(self):
        return tuple_prod(self.base_shape)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._shape()

    def forward(self, U) -> Union[FloatArray, IntArray, BoolArray]:
        return self._forward(U)

    def inverse(self, X) -> FloatArray:
        return self._inverse(X)

    def log_prob(self, X) -> FloatArray:
        log_prob = self._log_prob(X)
        if log_prob.size > 1:
            log_prob = jnp.sum(log_prob)
        if log_prob.shape != ():
            log_prob = log_prob.reshape(())
        return log_prob


class Prior(AbstractPrior):
    def __init__(self, dist_or_value: Union[tfpd.Distribution, jnp.ndarray], name: Optional[str] = None):
        super(Prior, self).__init__(name=name)
        if isinstance(dist_or_value, tfpd.Distribution):
            self._type = 'dist'
            self._dist = Distribution(dist_or_value)

        elif isinstance(dist_or_value, (jnp.ndarray, np.ndarray)):
            self._type = 'value'
            self._value = jnp.asarray(dist_or_value)
        self.name = name

    @property
    def dist(self) -> Distribution:
        if self._type != 'dist':
            raise ValueError(f"Wrong type, got {self._type}")
        return self._dist

    @property
    def value(self) -> jnp.ndarray:
        if self._type != 'value':
            raise ValueError(f"Wrong type, got {self._type}")
        return self._value

    def _base_shape(self) -> Tuple[int, ...]:
        if self._type == 'value':
            return ()
        elif self._type == 'dist':
            return self.dist.base_shape
        else:
            raise NotImplementedError()

    def _shape(self) -> Tuple[int, ...]:
        if self._type == 'value':
            return self.value.shape
        elif self._type == 'dist':
            return self.dist.shape
        else:
            raise NotImplementedError()

    def _dtype(self):
        if self._type == 'value':
            return self.value.dtype
        elif self._type == 'dist':
            return self.dist.dtype
        else:
            raise NotImplementedError()

    def _forward(self, U: FloatArray) -> Union[FloatArray, IntArray, BoolArray]:
        if self._type == 'value':
            return self.value
        elif self._type == 'dist':
            return self.dist.forward(U)
        else:
            raise NotImplementedError()

    def _inverse(self, X) -> FloatArray:
        if self._type == 'value':
            return jnp.asarray([], float_type)
        elif self._type == 'dist':
            return self.dist.inverse(X)
        else:
            raise NotImplementedError()

    def _log_prob(self, X: jnp.ndarray) -> FloatArray:
        if self._type == 'value':
            return jnp.asarray(0., float_type)
        elif self._type == 'dist':
            return self.dist.log_prob(X=X)
        else:
            raise NotImplementedError()


PriorModelGen = Generator[Prior, jnp.ndarray, Tuple[jnp.ndarray, ...]]
PriorModelType = Callable[[], PriorModelGen]


def parse_prior(prior_model: PriorModelType) -> Tuple[UType, XType]:
    """
    Computes placeholders of model.

    Args:
        prior_model: a callable that produces a prior model generator

    Returns:
        U placeholder, X placeholder
    """
    U_ndims = 0
    gen = prior_model()
    prior_response = None
    X_placeholder = dict()
    while True:
        try:
            prior: Prior = gen.send(prior_response)
            d = prior.base_ndims
            U_ndims += d
            u = jnp.zeros(prior.base_shape, float_type)
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in X_placeholder:
                    raise InvalidPriorName(name=prior.name)
                X_placeholder[prior.name] = prior_response
        except StopIteration:
            break
    U_placeholder = jnp.zeros((U_ndims,), float_type)
    return U_placeholder, X_placeholder


def transform(U: UType, prior_model: PriorModelType) -> XType:
    """
    Transforms a flat array of `U_ndims` i.i.d. samples of U[0,1] into the target prior.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        the prior variables
    """

    gen = prior_model()
    prior_response = None
    collection = dict()
    idx = 0
    while True:
        try:
            prior: Prior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
            if prior.name is not None:
                if prior.name in collection:
                    raise InvalidPriorName(name=prior.name)
                collection[prior.name] = prior_response
        except StopIteration:
            break
    return collection


def prepare_input(U: UType, prior_model: PriorModelType) -> LikelihoodInputType:
    """
    Transforms a flat array of `U_ndims` i.i.d. samples of U[0,1] into the likelihood conditional variables.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        the conditional variables of likelihood model
    """

    gen = prior_model()
    prior_response = None
    idx = 0
    while True:
        try:
            prior: Prior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
        except StopIteration as e:
            output = e.value
            if not isinstance(output, tuple):
                output = (output,)
            break
    return output


def log_prob_prior(U: UType, prior_model: PriorModelType) -> FloatArray:
    """
    Computes the prior log-density from a U-space sample.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator

    Returns:
        prior log-density
    """

    gen = prior_model()
    prior_response = None
    log_prob = []
    idx = 0
    while True:
        try:
            prior: Prior = gen.send(prior_response)
            d = prior.base_ndims
            u = jnp.reshape(U[idx:idx + d], prior.base_shape)
            idx += d
            prior_response = prior.forward(u)
            log_prob.append(prior.log_prob(prior_response))
        except StopIteration:
            break
    return sum(log_prob, jnp.asarray(0., float_type))


def compute_log_likelihood(U: UType, prior_model: PriorModelType, log_likelihood: LikelihoodType,
                           allow_nan: bool = False) -> FloatArray:
    """
    Computes the log likelihood from U-space sample.

    Args:
        U: [U_ndims] a flat array of i.i.d. samples of U[0,1]
        prior_model: a callable that produces a prior model generator
        log_likelihood: callable that takes arrays returned by the prior model and returns a scalar float
        allow_nan: whether to allow nans in likelihood

    Returns:
        log-likelihood
    """

    V = prepare_input(U=U, prior_model=prior_model)
    log_L = jnp.asarray(log_likelihood(*V), float_type)
    if not allow_nan:
        log_L = jnp.where(jnp.isnan(log_L), -jnp.inf, log_L)
    if log_L.size != 1:
        raise ValueError(f"Log likelihood should be scalar, but got {log_L.shape}.")
    if log_L.shape != ():
        log_L = jnp.reshape(log_L, ())
    return log_L
